#![no_std]

/*
heavily based upon https://www.geeksforgeeks.org/introduction-of-b-tree-2
 */

extern crate alloc;

use alloc::collections::VecDeque;
use alloc::vec;
use alloc::vec::Vec;

pub trait ToBytes {
    fn to_bytes(&self) -> Vec<u8>;
}

pub trait FromBytes {
    fn from_bytes(bytes: &[u8]) -> Self;
}

pub trait SizeOf {
    fn size_of(&self) -> u32;
}

impl<T: ToBytes> ToBytes for Option<T> {
    fn to_bytes(&self) -> Vec<u8> {
        match self {
            Some(x) => {
                let mut result = vec![1];
                result.extend_from_slice(&x.to_bytes());
                result
            }
            None => vec![0],
        }
    }
}

impl<T: FromBytes> FromBytes for Option<T> {
    fn from_bytes(bytes: &[u8]) -> Self {
        if bytes[0] == 0 {
            None
        } else {
            Some(T::from_bytes(&bytes[1..]))
        }
    }
}

impl ToBytes for u32 {
    fn to_bytes(&self) -> Vec<u8> {
        self.to_be_bytes().to_vec()
    }
}

impl FromBytes for u32 {
    fn from_bytes(bytes: &[u8]) -> Self {
        u32::from_be_bytes([bytes[0], bytes[1], bytes[2], bytes[3]])
    }
}

impl SizeOf for u32 {
    fn size_of(&self) -> u32 {
        4
    }
}


#[derive(Clone)]
pub struct BTreeNode<T: SizeOf + Clone + Default + ToBytes + FromBytes> {
    pub leaf: bool,
    pub keys: Vec<(u32, T)>,
    pub children: Vec<usize>,
    pub n: u32,
    pub m: u32,
}

#[derive(Clone)]
pub struct BTree<T: SizeOf + Clone + Default + ToBytes + FromBytes> {
    pub nodes: Vec<BTreeNode<T>>,
    pub root: Option<usize>,
    pub m: u32,
}

impl<T: SizeOf + Clone + Default + ToBytes + FromBytes> BTreeNode<T> {
    pub fn new(m: u32, leaf: bool) -> Self {
        BTreeNode {
            leaf,
            keys: vec![(0, T::default()); (2 * m - 1) as usize],
            children: vec![0; 2 * m as usize],
            n: 0,
            m,
        }
    }
}

// u8: leaf
// u16: keys length
// u16: children length
// u32: n
// [keys_length; (u32, sizeof(T) as u32, T)]: keys
// [children_length; u32]: children

impl<T: SizeOf + Clone + Default + ToBytes + FromBytes> ToBytes for BTreeNode<T> {
    fn to_bytes(&self) -> Vec<u8> {
        let mut result = vec![];
        result.push(self.m as u8);
        result.push(self.leaf as u8);
        result.extend_from_slice(&(self.keys.len() as u16).to_be_bytes());
        result.extend_from_slice(&(self.children.len() as u16).to_be_bytes());
        result.extend_from_slice(&self.n.to_be_bytes());
        for key in &self.keys {
            result.extend_from_slice(&key.0.to_be_bytes());
            result.extend_from_slice(&key.1.size_of().to_be_bytes());
            result.extend_from_slice(&key.1.to_bytes());
        }
        for child in &self.children {
            result.extend_from_slice(&(*child as u32).to_be_bytes());
        }
        result
    }
}

impl<T: SizeOf + Clone + Default + ToBytes + FromBytes> FromBytes for BTreeNode<T> {
    fn from_bytes(bytes: &[u8]) -> Self {
        let m = bytes[0] as u32;
        let mut result = BTreeNode::new(m, false);
        result.leaf = bytes[1] != 0;
        let keys_length = u16::from_be_bytes([bytes[2], bytes[3]]);
        let children_length = u16::from_be_bytes([bytes[4], bytes[5]]);
        result.n = u32::from_be_bytes([bytes[6], bytes[7], bytes[8], bytes[9]]);
        let mut i = 10;
        for j in 0..keys_length {
            let key = u32::from_be_bytes([bytes[i], bytes[i + 1], bytes[i + 2], bytes[i + 3]]);
            i += 4;
            let size = u32::from_be_bytes([bytes[i], bytes[i + 1], bytes[i + 2], bytes[i + 3]]);
            i += 4;
            let value = T::from_bytes(&bytes[i..i + size as usize]);
            i += size as usize;
            result.keys[j as usize] = (key, value);
        }
        for j in 0..children_length {
            let child = u32::from_be_bytes([bytes[i], bytes[i + 1], bytes[i + 2], bytes[i + 3]]);
            i += 4;
            result.children[j as usize] = child as usize;
        }
        result
    }
}

impl<T: SizeOf + Clone + Default + ToBytes + FromBytes> BTree<T> {
    pub fn new(m: u32) -> Self {
        BTree {
            nodes: vec![],
            root: None,
            m,
        }
    }

    /// deletes a node from the internal node list, and updates all references to the following nodes so that they point to the correct node
    pub fn delete_from_internal_list(&mut self, n: usize) {
        for node in &mut self.nodes {
            for child in &mut node.children {
                if *child >= n && *child > 0 {
                    *child -= 1;
                }
            }
        }
        self.nodes.remove(n);
    }

    /// finds the node with the given key, or None if not found
    pub fn search(&mut self, key: u32) -> Option<&mut BTreeNode<T>> {
        let mut index = self.root?;
        loop {
            let mut i = 0;
            while i < self.nodes[index].n as usize && self.nodes[index].keys[i].0 < key {
                i += 1;
            }
            if i < self.nodes[index].n as usize && self.nodes[index].keys[i].0 == key {
                return Some(&mut self.nodes[index]);
            }
            if self.nodes[index].leaf {
                return None;
            }
            index = self.nodes[index].children[i];
        }
    }

    /// splits the child `y` of node `node`, assumed that the node must be full
    pub fn split_child(&mut self, i: usize, y: usize, node: usize) {
        let mut z = BTreeNode::new(self.m, self.nodes[y].leaf);
        z.n = self.m - 1;

        // copy the last (m-1) keys of y into new_node
        for j in 0..self.m - 1 {
            z.keys[j as usize] = self.nodes[y].keys[(j + self.m) as usize].clone();
        }

        // copy the last m children of y into new_node
        if !self.nodes[y].leaf {
            for j in 0..self.m {
                z.children[j as usize] = self.nodes[y].children[(j + self.m) as usize];
            }
        }

        self.nodes[y].n = self.m - 1;

        // canonicalize node
        self.nodes.push(z);
        let z = self.nodes.len() - 1;

        // insert new_node into node
        let mut j = self.nodes[node].n;
        while j >= (i + 1) as u32 {
            self.nodes[node].children[j as usize + 1] = self.nodes[node].children[j as usize];
            j -= 1;
        }

        self.nodes[node].children[i + 1] = z;

        // a key of y will move to this node, so we need to make room for it
        let mut j = self.nodes[node].n as i32 - 1;
        while j >= i as i32 {
            self.nodes[node].keys[j as usize + 1] = self.nodes[node].keys[j as usize].clone();
            j -= 1;
        }

        // copy the middle key of y into node
        self.nodes[node].keys[i] = self.nodes[y].keys[(self.m - 1) as usize].clone();

        self.nodes[node].n += 1;
    }

    /// inserts the given key into the tree, assuming that the containing node is not full
    pub fn insert_assuming_not_full(&mut self, key: u32, value: T, node: usize) {
        let mut node = node;

        loop {
            let mut i = (self.nodes[node].n - 1) as isize;
            if self.nodes[node].leaf {
                while i >= 0 && self.nodes[node].keys[i as usize].0 > key {
                    self.nodes[node].keys[i as usize + 1] = self.nodes[node].keys[i as usize].clone();
                    i -= 1;
                }

                self.nodes[node].keys[(i + 1) as usize] = (key, value);
                self.nodes[node].n += 1;
                return;
            } else {
                // find the child which is going to contain the key
                while i >= 0 && self.nodes[node].keys[i as usize].0 > key {
                    i -= 1;
                }

                // check if the child is full
                if self.nodes[self.nodes[node].children[(i + 1) as usize]].n == (2 * self.m as usize - 1) as u32 {
                    self.split_child((i + 1) as usize, self.nodes[node].children[(i + 1) as usize], node);

                    if self.nodes[node].keys[(i + 1) as usize].0 < key {
                        i += 1;
                    }
                }
                node = self.nodes[node].children[(i + 1) as usize];
            }
        }
    }

    /// inserts the given key into the tree
    pub fn insert(&mut self, key: u32, value: T) {
        if self.root.is_none() {
            self.root = Some(0);
            self.nodes.clear();
            let mut new = BTreeNode::new(self.m, true);
            new.keys[0] = (key, value);
            new.n = 1;
            self.nodes.push(new);
        } else if self.nodes[self.root.unwrap()].n >= (2 * self.m - 1) {
            let mut new = BTreeNode::new(self.m, false);
            let root = self.root.unwrap();
            new.children[0] = root;
            let new_index = self.nodes.len();
            self.nodes.push(new);
            self.split_child(0, root, new_index);

            // new root has two children now, decide which of the two will contain the key
            let mut i = 0;
            if self.nodes[new_index].keys[0].0 < key {
                i += 1;
            }
            self.insert_assuming_not_full(key, value, self.nodes[new_index].children[i]);

            self.root = Some(new_index);
        } else {
            self.insert_assuming_not_full(key, value, self.root.unwrap());
        }
    }

    /// removes the given key from this b-tree (THIS IS THE FUNCTION YOU LIKELY WANT)
    pub fn remove(&mut self, key: u32) {
        if let Some(mut root) = self.root {
            let mut res = self.remove_from_nodetree(root, key);
            let mut root_removed = false;
            while let Some(node) = res.pop_front() {
                if root >= node {
                    root -= 1;
                }
                if node == root {
                    root_removed = true;
                }
            }
            self.root = Some(root);
            // if root has no keys, make the first child the new root (if it has a child), else set root to None
            if self.nodes[root].n == 0 || root_removed {
                if self.nodes[root].leaf {
                    self.delete_from_internal_list(root);
                    self.root = None;
                } else {
                    self.root = Some(self.nodes[root].children[0]);
                    self.delete_from_internal_list(root);
                }
            }
        }
    }

    /// removes the given key from the tree starting from a given node, returns VecDeque<usize> where Vec<usize> is a vector of nodes that were removed (so that other references can be updated)
    pub fn remove_from_nodetree(&mut self, mut node: usize, key: u32) -> VecDeque<usize> {
        let index = self.find_gte_key_on_node(node, key);

        let mut result = VecDeque::new();

        if index < self.nodes[node].n as usize && self.nodes[node].keys[index].0 == key {
            // if this is a leaf node, call remove_assuming_leaf
            // vice versa
            if self.nodes[node].leaf {
                self.remove_assuming_leaf(node, index);
            } else {
                let res = self.remove_assuming_non_leaf(node, index);
                result.append(&mut res.clone());
            }

            result
        } else {
            // if this is a leaf node, key not found
            if self.nodes[node].leaf {
                return result;
            }

            // saved for later
            let flag = index == self.nodes[node].n as usize;

            // if the child that should contain the key has less than m keys, steal from it
            if self.nodes[self.nodes[node].children[index]].n < self.m {
                let shift = self.steal_give(node, index);
                if let Some(shift) = shift {
                    // if node is greater than shift, decrease them by 1
                    if node >= shift {
                        node -= 1;
                    }
                    result.push_front(shift);
                }
            }

            // if the last child was merged, recurse on the (index-1)th child
            // vice versa
            if flag && index > self.nodes[node].n as usize {
                let res = self.remove_from_nodetree(self.nodes[node].children[index - 1], key);
                result.append(&mut res.clone());
            } else {
                let res = self.remove_from_nodetree(self.nodes[node].children[index], key);
                result.append(&mut res.clone());
            }

            result
        }
    }

    /// finds the first key on a node that is greater than or equal to the given key
    pub fn find_gte_key_on_node(&mut self, node: usize, key: u32) -> usize {
        let mut i = 0;
        while i < self.nodes[node].n && self.nodes[node].keys[i as usize].0 < key {
            i += 1;
        }
        i as usize
    }

    /// steals keys from a node's children to fill child n, returns the position of the node that was removed so that other references can be updated, or None if no node was removed
    pub fn steal_give(&mut self, node: usize, n: usize) -> Option<usize> {
        // if the previous child has more than m-1 keys, steal from it
        // else if the next child has more than m-1 keys, steal from it
        // else, merge with one of the children
        if n != 0 && self.nodes[self.nodes[node].children[n-1]].n >= self.m {
            self.cfptn(node, n);
            None
        } else if n != self.nodes[node].n as usize && self.nodes[self.nodes[node].children[n+1]].n >= self.m {
            self.cfftn(node, n);
            None
        } else if n != self.nodes[node].n as usize {
            Some(self.merge_children(node, n).unwrap())
        } else {
            Some(self.merge_children(node, n-1).unwrap())
        }
    }

    /// copies the last key from the (n+1)th child and gives it to the nth child
    pub fn cfftn(&mut self, node: usize, n: usize) {
        let child = self.nodes[node].children[n];
        let sibling = self.nodes[node].children[n+1];

        // give child the nth key from the current node
        let child_n = self.nodes[child].n;
        self.nodes[child].keys[child_n as usize] = self.nodes[node].keys[n].clone();

        // if child isn't leaf, move sibling's first child as child's last child
        if !self.nodes[child].leaf {
            self.nodes[child].children[child_n as usize + 1] = self.nodes[sibling].children[0];
        }

        // move the first key from sibling to the current node
        self.nodes[node].keys[n] = self.nodes[sibling].keys[0].clone();

        // shift sibling keys to the left
        for i in 1..self.nodes[sibling].n {
            self.nodes[sibling].keys[i as usize - 1] = self.nodes[sibling].keys[i as usize].clone();
        }

        // if sibling isn't leaf, shift sibling children to the left
        if !self.nodes[sibling].leaf {
            for i in 1..self.nodes[sibling].n+1 {
                self.nodes[sibling].children[i as usize - 1] = self.nodes[sibling].children[i as usize];
            }
        }

        // decrease number of keys in sibling
        self.nodes[sibling].n -= 1;
        // increase number of keys in child
        self.nodes[child].n += 1;
    }

    /// copies the last key from the (n-1)th child and gives it to the nth child
    pub fn cfptn(&mut self, node: usize, n: usize) {
        let child = self.nodes[node].children[n];
        let sibling = self.nodes[node].children[n-1];

        // move all keys in child one to the right
        for i in (0..=self.nodes[child].n-1).rev() {
            self.nodes[child].keys[i as usize + 1] = self.nodes[child].keys[i as usize].clone();
        }

        // if child isn't leaf, move all child pointers one step ahead
        if !self.nodes[child].leaf {
            for i in (0..=self.nodes[child].n).rev() {
                self.nodes[child].children[i as usize + 1] = self.nodes[child].children[i as usize];
            }
        }

        // give child the (n-1)th key from the current node
        self.nodes[child].keys[0] = self.nodes[node].keys[n-1].clone();

        // if child isn't leaf, move sibling's last child as child's first child
        if !self.nodes[child].leaf {
            self.nodes[child].children[0] = self.nodes[sibling].children[self.nodes[sibling].n as usize];
        }

        // move the (n-1)th key from sibling to the current node
        self.nodes[node].keys[n-1] = self.nodes[sibling].keys[self.nodes[sibling].n as usize - 1].clone();

        // decrease number of keys in sibling
        self.nodes[sibling].n -= 1;
        // increase number of keys in child
        self.nodes[child].n += 1;
    }

    /// merges the given node's nth child with its (n+1)th child, returns Some(n) if the merge was successful, None if the merge was not successful
    /// where `n` is the index of the node that was removed so that other references can be updated
    pub fn merge_children(&mut self, node: usize, n: usize) -> Option<usize> {
        let child = self.nodes[node].children[n];
        let sibling = self.nodes[node].children[n+1];

        // pop key from current node and insert it into child
        self.nodes[child].keys[self.m as usize - 1] = self.nodes[node].keys[n].clone();

        // copy keys from sibling into child
        for i in 0..self.nodes[sibling].n {
            self.nodes[child].keys[(self.m + i) as usize] = self.nodes[sibling].keys[i as usize].clone();
        }

        // if child isn't leaf, copy children from sibling into child
        if !self.nodes[child].leaf {
            for i in 0..self.nodes[sibling].n {
                self.nodes[child].children[(self.m + i) as usize] = self.nodes[sibling].children[i as usize];
            }
        }

        // move keys after n one to the left, filling the gap
        for i in n+1..self.nodes[node].n as usize {
            self.nodes[node].keys[i-1] = self.nodes[node].keys[i].clone();
        }

        // move children after n+1 one to the left, filling the gap
        for i in n+2..self.nodes[node].n as usize + 1 {
            self.nodes[node].children[i-1] = self.nodes[node].children[i];
        }
        // decrease number of keys in node
        self.nodes[child].n += self.nodes[sibling].n + 1;
        self.nodes[node].n -= 1;

        // delete sibling
        self.delete_from_internal_list(sibling);

        Some(sibling)
    }

    /// removes the nth key from the given node assuming it is a non-leaf node, returns VecDeque<usize> where Vec<usize> is a vector of nodes that were removed (so that other references can be updated)
    pub fn remove_assuming_non_leaf(&mut self, mut node: usize, n: usize) -> VecDeque<usize> {
        let key = self.nodes[node].keys[n].0;

        let mut result = VecDeque::new();

        let child_n = self.nodes[self.nodes[node].children[n]].n;

        // if child before k (children[n]) has at least m keys, find predecessor, replace key with predecessor, delete predecessor
        // else if nth child has less than m keys, and (n+1)th child has at least m keys, replace key with successor, delete successor
        // else merge nth child with (n+1)th child, recurse on nth child
        if child_n >= self.m {
            let pre_node = self.fkfl(node, n).unwrap().clone();
            self.nodes[node].keys[n] = pre_node.clone();
            let res = self.remove_from_nodetree(self.nodes[node].children[n], pre_node.0);
            result.append(&mut res.clone());
        } else if self.nodes[self.nodes[node].children[n+1]].n >= self.m {
            let suc_node = self.fkfr(node, n).unwrap().clone();
            self.nodes[node].keys[n] = suc_node.clone();
            let res = self.remove_from_nodetree(self.nodes[node].children[n+1], suc_node.0);
            result.append(&mut res.clone());
        } else {
            let res = self.merge_children(node, n).unwrap();
            if node >= res {
                node -= 1;
                result.push_front(res);
            }
            let res = self.remove_from_nodetree(self.nodes[node].children[n], key);
            result.append(&mut res.clone());
        }

        result
    }

    /// removes the nth key from the given node assuming it is a leaf node
    pub fn remove_assuming_leaf(&mut self, node: usize, n: usize) {
        // move all keys after n one to the left
        for i in n+1..self.nodes[node].n as usize {
            self.nodes[node].keys[i-1] = self.nodes[node].keys[i].clone();
        }
        self.nodes[node].n -= 1;
    }

    /// returns the last key of the last leaf node starting from the right of the children of the given node
    pub fn fkfr(&self, node: usize, i: usize) -> Option<&(u32, T)> {
        let children = &self.nodes[node].children;
        if children.is_empty() {
            return None;
        }
        let mut node = children[i];
        while !self.nodes[node].leaf {
            node = self.nodes[node].children[self.nodes[node].n as usize];
        }
        Some(&self.nodes[node].keys[self.nodes[node].n as usize - 1])
    }

    /// returns the first key of the first leaf node starting from the left of the children of the given node
    pub fn fkfl(&self, node: usize, i: usize) -> Option<&(u32, T)> {
        let children = &self.nodes[node].children;
        if children.is_empty() {
            return None;
        }
        let mut node = children[i];
        while !self.nodes[node].leaf {
            node = self.nodes[node].children[0];
        }
        Some(&self.nodes[node].keys[0])

    }

    /// traverses the tree in order and returns a vector of all keys
    pub fn traverse_in_order(&self, node: usize) -> Vec<u32> {
        let mut result = vec![];
        for i in 0..self.nodes[node].n {
            if !self.nodes[node].leaf {
                result.append(&mut self.traverse_in_order(self.nodes[node].children[i as usize]));
            }
            result.push(self.nodes[node].keys[i as usize].0);
        }

        if !self.nodes[node].leaf {
            result.append(&mut self.traverse_in_order(self.nodes[node].children[self.nodes[node].n as usize]));
        }

        result
    }

    /// traverses the tree in order and returns a vector of all values {
    pub fn traverse_in_order_values(&self, node: usize) -> Vec<T> {
        let mut result = vec![];
        for i in 0..self.nodes[node].n {
            if !self.nodes[node].leaf {
                result.append(&mut self.traverse_in_order_values(self.nodes[node].children[i as usize]));
            }
            result.push(self.nodes[node].keys[i as usize].1.clone());
        }

        if !self.nodes[node].leaf {
            result.append(&mut self.traverse_in_order_values(self.nodes[node].children[self.nodes[node].n as usize]));
        }

        result
    }
}

impl<T: SizeOf + Clone + Default + ToBytes + FromBytes> ToBytes for BTree<T> {
    fn to_bytes(&self) -> Vec<u8> {
        let mut result = vec![];
        result.extend_from_slice(&self.root.map(|u| u as u32).to_bytes());
        result.extend_from_slice(&self.m.to_be_bytes());
        result.extend_from_slice(&(self.nodes.len() as u32).to_be_bytes());
        for node in &self.nodes {
            let bytes = node.to_bytes();
            result.extend_from_slice(&(bytes.len() as u32).to_be_bytes());
            result.extend_from_slice(&bytes);
        }
        result
    }
}

impl<T: SizeOf + Clone + Default + ToBytes + FromBytes> FromBytes for BTree<T> {
    fn from_bytes(bytes: &[u8]) -> Self {
        let root = Option::<u32>::from_bytes(&bytes[0..5]).map(|u| u as usize);
        let m = u32::from_be_bytes([bytes[5], bytes[6], bytes[7], bytes[8]]);
        let nodes_length = u32::from_be_bytes([bytes[9], bytes[10], bytes[11], bytes[12]]);
        let mut nodes = vec![];
        let mut i = 13;
        for _ in 0..nodes_length {
            let node_length = u32::from_be_bytes([bytes[i], bytes[i + 1], bytes[i + 2], bytes[i + 3]]);
            i += 4;
            nodes.push(BTreeNode::<T>::from_bytes(&bytes[i..i + node_length as usize]));
            i += node_length as usize;
        }
        BTree {
            nodes,
            root,
            m,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn it_works() {
        impl ToBytes for i32 {
            fn to_bytes(&self) -> Vec<u8> {
                self.to_be_bytes().to_vec()
            }
        }

        impl FromBytes for i32 {
            fn from_bytes(bytes: &[u8]) -> Self {
                i32::from_be_bytes([bytes[0], bytes[1], bytes[2], bytes[3]])
            }
        }

        impl SizeOf for i32 {
            fn size_of(&self) -> u32 {
                4
            }
        }

        let mut tree = BTree::new(3);
        tree.insert(10, 0);
        tree.insert(20, 0);
        tree.insert(5, 0);
        tree.insert(6, 1);
        tree.insert(12, 1);
        tree.insert(30, 0);
        tree.insert(7, 0);
        tree.insert(17, 0);
        tree.insert(11, 0);
        tree.insert(3, 0);
        tree.insert(2, 0);
        tree.insert(1, 0);
        tree.insert(24, 1);
        tree.insert(25, 0);

        tree.remove(25);
        tree.remove(24);

        fn tree_good(mut tree: BTree<i32>) {
            // print traversed tree
            assert_eq!(tree.traverse_in_order(tree.root.unwrap()), vec![1, 2, 3, 5, 6, 7, 10, 11, 12, 17, 20, 30]);
            // get values of 6, 12, 17
            assert_eq!(tree.search(6).map(|x| x.keys.iter().filter(|x| x.0 == 6).map(|x| x.1).collect::<Vec<_>>()), Some(vec![1]));
            assert_eq!(tree.search(12).map(|x| x.keys.iter().filter(|x| x.0 == 12).map(|x| x.1).collect::<Vec<_>>()), Some(vec![1]));
            assert_eq!(tree.search(17).map(|x| x.keys.iter().filter(|x| x.0 == 17).map(|x| x.1).collect::<Vec<_>>()), Some(vec![0]));
        }

        tree_good(tree.clone());

        // serialize and deserialize
        let bytes = tree.to_bytes();
        let tree2 = BTree::<i32>::from_bytes(&bytes);

        tree_good(tree2.clone());

        // remove 10 and 20, then check if tree is still good
        tree.remove(10);
        assert_eq!(tree.traverse_in_order(tree.root.unwrap()), vec![1, 2, 3, 5, 6, 7, 11, 12, 17, 20, 30]);
        tree.remove(20);
        assert_eq!(tree.traverse_in_order(tree.root.unwrap()), vec![1, 2, 3, 5, 6, 7, 11, 12, 17, 30]);

        // remove all keys but 30
        let keys_remaining = vec![1, 2, 3, 5, 6, 7, 11, 12, 17];
        for key in keys_remaining {
            tree.remove(key);
        }

        // check if tree is still good
        assert_eq!(tree.traverse_in_order(tree.root.unwrap()), vec![30]);

        // remove 30
        tree.remove(30);

        assert_eq!(tree.root, None);
        assert_eq!(tree.nodes.len(), 0);
    }
}
